"""Test windspharm computations against reference solutions."""

from __future__ import absolute_import

import numpy as np
import pytest

from skyborn.windspharm.tests import VectorWindTest, solvers

from .reference import reference_solutions


class SolutionTest(VectorWindTest):
    """Base class for all solution test classes."""

    interface = None
    gridtype = None
    radius = None
    legfunc = None

    @classmethod
    def setup_class(cls):
        msg = "missing dependencies required to test the {!s} interface"
        try:
            cls.solution = reference_solutions(cls.interface, cls.gridtype)
        except ValueError:
            pytest.skip(msg.format(cls.interface))
        cls.pre_modify_solution()
        try:
            # gridtype argument only available for the standard interface
            kwargs = {}
            if cls.interface == "standard":
                kwargs["gridtype"] = cls.gridtype
            if cls.radius is not None:
                kwargs["rsphere"] = cls.radius
            if cls.legfunc is not None:
                kwargs["legfunc"] = cls.legfunc
            cls.vw = solvers[cls.interface](
                cls.solution["uwnd"], cls.solution["vwnd"], **kwargs
            )
        except KeyError:
            pytest.skip(msg.format(cls.interface))
        cls.post_modify_solution()

    @classmethod
    def pre_modify_solution(cls):
        pass

    @classmethod
    def post_modify_solution(cls):
        pass

    def test_magnitude(self):
        # computed magnitude matches magnitude of reference solution?
        mag1 = self.vw.magnitude()
        mag2 = (self.solution["uwnd"] ** 2 + self.solution["vwnd"] ** 2) ** 0.5
        self.assert_error_is_zero(mag1, mag2)

    def test_vorticity(self):
        # computed vorticity matches reference solution?
        vrt1 = self.vw.vorticity()
        vrt2 = self.solution["vrt"]
        self.assert_error_is_zero(vrt1, vrt2)

    def test_divergence(self):
        # computed divergence matches reference solution?
        div1 = self.vw.divergence()
        div2 = self.solution["div"]
        self.assert_error_is_zero(div1, div2)

    def test_streamfunction(self):
        # computed streamfunction matches reference solution?
        sf1 = self.vw.streamfunction()
        sf2 = self.solution["psi"].copy()
        self.assert_error_is_zero(sf1, sf2)

    def test_velocitypotential(self):
        # computed velocity potential matches reference solution?
        vp1 = self.vw.velocitypotential()
        vp2 = self.solution["chi"].copy()
        self.assert_error_is_zero(vp1, vp2)

    def test_nondivergent(self):
        # computed non-divergent vector wind matches reference solution?
        upsi1, vpsi1 = self.vw.nondivergentcomponent()
        upsi2, vpsi2 = self.solution["upsi"], self.solution["vpsi"]
        self.assert_error_is_zero(upsi1, upsi2)
        self.assert_error_is_zero(vpsi1, vpsi2)

    def test_irrotational(self):
        # computed irrotational vector wind matches reference solution?
        uchi1, vchi1 = self.vw.irrotationalcomponent()
        uchi2, vchi2 = self.solution["uchi"], self.solution["vchi"]
        self.assert_error_is_zero(uchi1, uchi2)
        self.assert_error_is_zero(vchi1, vchi2)

    def test_gradient(self):
        # computed gradient matches reference solution?
        uchi1, vchi1 = self.vw.gradient(self.solution["chi"])
        uchi2, vchi2 = self.solution["chigradu"], self.solution["chigradv"]
        self.assert_error_is_zero(uchi1, uchi2)
        self.assert_error_is_zero(vchi1, vchi2)

    def test_vrtdiv(self):
        # vrtdiv() matches vorticity()/divergence()?
        vrt1, div1 = self.vw.vrtdiv()
        vrt2 = self.vw.vorticity()
        div2 = self.vw.divergence()
        self.assert_error_is_zero(vrt1, vrt2)
        self.assert_error_is_zero(div1, div2)

    def test_sfvp(self):
        # sfvp() matches streamfunction()/velocitypotential()?
        sf1, vp1 = self.vw.sfvp()
        sf2 = self.vw.streamfunction()
        vp2 = self.vw.velocitypotential()
        self.assert_error_is_zero(sf1, sf2)
        self.assert_error_is_zero(vp1, vp2)

    def test_helmholtz(self):
        # helmholtz() matches irrotationalcomponent()/nondivergentcomponent()?
        uchi1, vchi1, upsi1, vpsi1 = self.vw.helmholtz()
        uchi2, vchi2 = self.vw.irrotationalcomponent()
        upsi2, vpsi2 = self.vw.nondivergentcomponent()
        uchi, vchi, upsi2, vpsi2 = self.vw.helmholtz()
        self.assert_error_is_zero(uchi1, uchi2)
        self.assert_error_is_zero(vchi1, vchi2)
        self.assert_error_is_zero(upsi1, upsi2)
        self.assert_error_is_zero(vpsi1, vpsi2)

    def test_truncate(self):
        # vorticity truncated to T21 matches reference?
        vrt_trunc = self.vw.truncate(self.solution["vrt"], truncation=21)
        self.assert_error_is_zero(vrt_trunc, self.solution["vrt_trunc"])


# ----------------------------------------------------------------------------
# Tests for the standard interface


class StandardSolutionTest(SolutionTest):
    """Base class for all standard interface solution test classes."""

    interface = "standard"


class TestStandardRegular(StandardSolutionTest):
    """Regular grid."""

    gridtype = "regular"


class TestStandardGaussian(StandardSolutionTest):
    """Gaussian grid."""

    gridtype = "gaussian"


class TestStandardRegularSingleton(StandardSolutionTest):
    """Singleton right-most dimension."""

    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in cls.solution:
            cls.solution[field_name] = cls.solution[field_name][..., np.newaxis]


class TestStandardGaussianSingleton(StandardSolutionTest):
    """Singleton right-most dimension."""

    gridtype = "gaussian"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in cls.solution:
            cls.solution[field_name] = cls.solution[field_name][..., np.newaxis]


class TestStandardMultiTime(StandardSolutionTest):
    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in cls.solution:
            cls.solution[field_name] = cls.solution[field_name][..., np.newaxis].repeat(
                5, axis=-1
            )


class TestStandardRadiusDefaultExplicit(StandardSolutionTest):
    gridtype = "regular"
    radius = 6.3712e6


class TestStandardRadius(StandardSolutionTest):
    gridtype = "regular"
    radius = 6.3712e6 / 16.0

    @classmethod
    def post_modify_solution(cls):
        # Divergence and vorticity should be scaled by the inverse of the
        # radius factor.
        cls.solution["vrt"] = cls.solution["vrt"] * 16.0
        cls.solution["div"] = cls.solution["div"] * 16.0
        cls.solution["vrt_trunc"] = cls.solution["vrt_trunc"] * 16
        # Stream function and velocity potential should be scaled by the
        # radius factor.
        cls.solution["psi"] = cls.solution["psi"] / 16
        cls.solution["chi"] = cls.solution["chi"] / 16


class TestStandardLegfuncComputed(StandardSolutionTest):
    """Computed Legendre functions."""

    gridtype = "regular"
    legfunc = "computed"


# ----------------------------------------------------------------------------
# Tests for the Iris interface


class IrisSolutionTest(SolutionTest):
    """Base class for all Iris interface solution test classes."""

    interface = "iris"

    def test_truncate_reversed(self):
        vrt_trunc = self.vw.truncate(self.solution["vrt"][::-1], truncation=21)
        self.assert_error_is_zero(vrt_trunc, self.solution["vrt_trunc"])


class TestIrisRegular(IrisSolutionTest):
    """Regular grid."""

    gridtype = "regular"


class TestIrisGaussian(IrisSolutionTest):
    """Gaussian grid."""

    gridtype = "gaussian"


class TestIrisGridTranspose(IrisSolutionTest):
    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in cls.solution.keys():
            cls.solution[field_name].transpose([1, 0])

    def test_truncate_reversed(self):
        vrt_trunc = self.vw.truncate(self.solution["vrt"][:, ::-1], truncation=21)
        self.assert_error_is_zero(vrt_trunc, self.solution["vrt_trunc"])


class TestIrisInvertedLatitude(IrisSolutionTest):
    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in ("uwnd", "vwnd"):
            cls.solution[field_name] = cls.solution[field_name][::-1]

    @classmethod
    def post_modify_solution(cls):
        for field_name in ("uwnd", "vwnd"):
            cls.solution[field_name] = cls.solution[field_name][::-1]


class TestIrisRadiusDefaultExplicit(IrisSolutionTest):
    gridtype = "regular"
    radius = 6.3712e6


class TestIrisRadius(IrisSolutionTest):
    gridtype = "regular"
    radius = 6.3712e6 / 16.0

    @classmethod
    def post_modify_solution(cls):
        # Divergence and vorticity should be scaled by the inverse of the
        # radius factor.
        cls.solution["vrt"] = cls.solution["vrt"] * 16.0
        cls.solution["div"] = cls.solution["div"] * 16.0
        cls.solution["vrt_trunc"] = cls.solution["vrt_trunc"] * 16
        # Stream function and velocity potential should be scaled by the
        # radius factor.
        cls.solution["psi"] = cls.solution["psi"] / 16
        cls.solution["chi"] = cls.solution["chi"] / 16


class TestIrisLegfuncComputed(IrisSolutionTest):
    """Computed Legendre functions."""

    gridtype = "regular"
    legfunc = "computed"


# ----------------------------------------------------------------------------
# Tests for the Xarray interface


class XarraySolutionTest(SolutionTest):
    """Base class for all Xarray interface solution test classes."""

    interface = "xarray"

    def test_truncate_reversed(self):
        vrt_trunc = self.vw.truncate(self.solution["vrt"][::-1], truncation=21)
        self.assert_error_is_zero(vrt_trunc, self.solution["vrt_trunc"])


class TestXarrayRegular(XarraySolutionTest):
    """Regular grid."""

    gridtype = "regular"


class TestXarrayGaussian(XarraySolutionTest):
    """Gaussian grid."""

    gridtype = "gaussian"


class TestXarrayGridTranspose(XarraySolutionTest):
    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in cls.solution.keys():
            cls.solution[field_name] = cls.solution[field_name].transpose()

    def test_truncate_reversed(self):
        vrt_trunc = self.vw.truncate(self.solution["vrt"][:, ::-1], truncation=21)
        self.assert_error_is_zero(vrt_trunc, self.solution["vrt_trunc"])


class TestXarrayInvertedLatitude(XarraySolutionTest):
    gridtype = "regular"

    @classmethod
    def pre_modify_solution(cls):
        for field_name in ("uwnd", "vwnd"):
            cls.solution[field_name] = cls.solution[field_name][::-1]

    @classmethod
    def post_modify_solution(cls):
        for field_name in ("uwnd", "vwnd"):
            cls.solution[field_name] = cls.solution[field_name][::-1]


class TestXarrayRadiusDefaultExplicit(XarraySolutionTest):
    gridtype = "regular"
    radius = 6.3712e6


class TestXarrayRadius(XarraySolutionTest):
    gridtype = "regular"
    radius = 6.3712e6 / 16.0

    @classmethod
    def post_modify_solution(cls):
        # Divergence and vorticity should be scaled by the inverse of the
        # radius factor.
        cls.solution["vrt"] = cls.solution["vrt"] * 16.0
        cls.solution["div"] = cls.solution["div"] * 16.0
        cls.solution["vrt_trunc"] = cls.solution["vrt_trunc"] * 16
        # Stream function and velocity potential should be scaled by the
        # radius factor.
        cls.solution["psi"] = cls.solution["psi"] / 16
        cls.solution["chi"] = cls.solution["chi"] / 16


class TestXarrayLegfuncComputed(XarraySolutionTest):
    """Computed Legendre functions."""

    gridtype = "regular"
    legfunc = "computed"



from typing import Optional, Union

import numpy as np

__all__ = [
    "calculate_gradient",
    "calculate_meridional_gradient",
    "calculate_zonal_gradient",
    "calculate_vertical_gradient",
    "EARTH_RADIUS",
]

EARTH_RADIUS = 6371e3  # Earth's mean radius (m)


def calculate_gradient(
    field: np.ndarray,
    coordinates: np.ndarray,
    axis: int = -1,
    radius: Optional[float] = None,
) -> np.ndarray:
    """Calculate gradient of an arbitrary dimensional array along specified coordinates

    Args:
        field: Data field for gradient calculation, can be any dimensional array, e.g., (time, level, lat, lon)
        coordinates: Coordinate array along which to calculate the gradient, such as latitude or longitude values
        axis: Specifies the dimensional axis for gradient calculation, defaults to the last dimension (-1)
        radius: Earth radius, default is None. If provided, coordinates are treated as latitude in degrees
                and converted to distance in meters. If None, coordinates are used directly.

    Returns:
        Gradient field with the same shape as the input field
    """
    # Check if input data dimensions match
    if coordinates.size != field.shape[axis]:
        raise ValueError(
            f"Coordinate array size ({coordinates.size}) does not match field size ({field.shape[axis]}) on the specified axis"
        )

    # If radius is provided, treat coordinates as latitude and convert to distance
    if radius is not None:
        # Convert latitude to actual distance
        distances = coordinates * np.pi / 180.0 * radius
    else:
        # Use coordinates directly
        distances = coordinates

    # Create output array with the same shape as input
    gradient = np.zeros_like(field, dtype=float)

    # Handle edge case: single point (no gradient can be computed)
    if field.shape[axis] < 2:
        return gradient  # Return zero gradient

    # To use numpy's advanced indexing, we need to create index arrays
    ndim = field.ndim
    idx_ranges = [slice(None)] * ndim

    # Handle the simple case where distances are uniform
    dx = np.diff(distances)
    if len(dx) > 0 and np.allclose(dx, dx[0]):
        # Uniform spacing - use simple central differences
        h = dx[0]

        # Central differences for interior points
        for i in range(1, field.shape[axis] - 1):
            idx = idx_ranges.copy()
            idx[axis] = i
            idx_forward = idx_ranges.copy()
            idx_forward[axis] = i + 1
            idx_backward = idx_ranges.copy()
            idx_backward[axis] = i - 1

            gradient[tuple(idx)] = (
                field[tuple(idx_forward)] - field[tuple(idx_backward)]
            ) / (2.0 * h)

        # Forward difference for left boundary
        idx_left = idx_ranges.copy()
        idx_left[axis] = 0
        idx_left_plus = idx_ranges.copy()
        idx_left_plus[axis] = 1
        gradient[tuple(idx_left)] = (
            field[tuple(idx_left_plus)] - field[tuple(idx_left)]
        ) / h

        # Backward difference for right boundary
        idx_right = idx_ranges.copy()
        idx_right[axis] = -1
        idx_right_minus = idx_ranges.copy()
        idx_right_minus[axis] = -2
        gradient[tuple(idx_right)] = (
            field[tuple(idx_right)] - field[tuple(idx_right_minus)]
        ) / h

    else:
        # Non-uniform spacing - use variable spacing formulas
        for i in range(field.shape[axis]):
            idx = idx_ranges.copy()
            idx[axis] = i

            if i == 0:
                # Forward difference for first point
                idx_plus = idx_ranges.copy()
                idx_plus[axis] = 1
                gradient[tuple(idx)] = (field[tuple(idx_plus)] - field[tuple(idx)]) / (
                    distances[1] - distances[0]
                )
            elif i == field.shape[axis] - 1:
                # Backward difference for last point
                idx_minus = idx_ranges.copy()
                idx_minus[axis] = i - 1
                gradient[tuple(idx)] = (field[tuple(idx)] - field[tuple(idx_minus)]) / (
                    distances[i] - distances[i - 1]
                )
            else:
                # Central difference for interior points
                idx_forward = idx_ranges.copy()
                idx_forward[axis] = i + 1
                idx_backward = idx_ranges.copy()
                idx_backward[axis] = i - 1

                h1 = distances[i] - distances[i - 1]
                h2 = distances[i + 1] - distances[i]

                # Use weighted central difference for non-uniform spacing
                gradient[tuple(idx)] = (
                    -h2 / (h1 * (h1 + h2)) * field[tuple(idx_backward)]
                    + (h2 - h1) / (h1 * h2) * field[tuple(idx)]
                    + h1 / (h2 * (h1 + h2)) * field[tuple(idx_forward)]
                )

    return gradient


def calculate_meridional_gradient(
    field: np.ndarray,
    latitudes: np.ndarray,
    lat_axis: int = -1,
    radius: float = 6371000.0,
) -> np.ndarray:
    """Calculate meridional gradient (gradient along latitude direction)

    Args:
        field: Data field for gradient calculation, can be any dimensional array
        latitudes: Latitude array (degrees)
        lat_axis: Specifies the axis for latitude, defaults to the last dimension (-1)
        radius: Earth radius, default is 6371000.0 meters

    Returns:
        Meridional gradient field
    """
    return calculate_gradient(field, latitudes, axis=lat_axis, radius=radius)


def calculate_vertical_gradient(
    field: np.ndarray, pressure: np.ndarray, pressure_axis: int = -3
) -> np.ndarray:
    """Calculate vertical gradient (gradient along pressure direction)

    Args:
        field: Data field for gradient calculation
        pressure: Pressure array (Pa), must be monotonically decreasing
        pressure_axis: Specifies the axis for pressure, defaults to the third-to-last dimension (-3)

    Returns:
        Vertical gradient field
    """
    return calculate_gradient(field, pressure, axis=pressure_axis, radius=None)


def calculate_zonal_gradient(
    field: np.ndarray,
    longitudes: np.ndarray,
    latitudes: np.ndarray,
    lon_axis: int = -1,
    lat_axis: int = -2,
    radius: float = 6371000.0,
) -> np.ndarray:
    """Calculate zonal gradient (gradient along longitude direction)

    Args:
        field: Data field for gradient calculation, can be any dimensional array
        longitudes: Longitude array (degrees)
        latitudes: Latitude array (degrees), used to calculate actual distance between longitudes at different latitudes
        lon_axis: Specifies the axis for longitude, defaults to the last dimension (-1)
        lat_axis: Specifies the axis for latitude, defaults to the second-to-last dimension (-2)
        radius: Earth radius, default is 6371000.0 meters

    Returns:
        Zonal gradient field
    """
    # Get latitude factor to adjust actual distance between longitudes at different latitudes
    cos_lat = np.cos(np.radians(latitudes))

    # If field is 4D (time, level, lat, lon)
    if field.ndim == 4 and lon_axis == -1 and lat_axis == -2:
        # Create a latitude factor array with shape suitable for broadcasting
        cos_lat_expanded = cos_lat.reshape(1, 1, -1, 1)

        # Convert longitudes to actual distances considering latitude
        effective_distances = np.radians(longitudes) * radius * cos_lat_expanded

        # Calculate gradient
        return calculate_gradient(field, effective_distances, axis=lon_axis, radius=1.0)

    # If field is 3D (time, lat, lon)
    elif field.ndim == 3 and lon_axis == -1 and lat_axis == -2:
        cos_lat_expanded = cos_lat.reshape(1, -1, 1)
        effective_distances = np.radians(longitudes) * radius * cos_lat_expanded
        return calculate_gradient(field, effective_distances, axis=lon_axis, radius=1.0)

    else:
        # For other dimension combinations, create appropriate broadcasting shape
        broadcast_shape = [1] * field.ndim
        broadcast_shape[lat_axis] = len(latitudes)
        cos_lat_expanded = cos_lat.reshape(broadcast_shape)

        # Create effective distance array
        effective_longitudes = np.radians(longitudes) * radius

        # Calculate gradient for each latitude
        result = np.zeros_like(field)

        # Loop through each latitude (implementation depends on specific data structure, may need adjustment)
        for i in range(len(latitudes)):
            idx = [slice(None)] * field.ndim
            idx[lat_axis] = i

            # Adjust longitude distance for current latitude
            current_effective_dist = effective_longitudes * cos_lat[i]

            # Calculate gradient for current latitude
            result[tuple(idx)] = calculate_gradient(
                field[tuple(idx)], current_effective_dist, axis=lon_axis, radius=1.0
            )

        return result



"""
Tests for skyborn.plot.plotting module.

This module tests the plotting functionality in skyborn.plot.plotting,
including matplotlib integration and image comparison tests.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pytest

from skyborn.plot.plotting import add_equal_axes, createFigure


class TestCreateFigure:
    """Test the createFigure function."""

    def test_createFigure_basic(self):
        """Test basic createFigure functionality."""
        fig = createFigure()

        # Check default parameters
        assert fig.get_figwidth() == 12
        assert fig.get_figheight() == 8
        assert fig.dpi == 300

        plt.close(fig)

    def test_createFigure_custom_figsize(self):
        """Test createFigure with custom figsize."""
        fig = createFigure(figsize=(8, 6))

        assert fig.get_figwidth() == 8
        assert fig.get_figheight() == 6
        assert fig.dpi == 300

        plt.close(fig)

    def test_createFigure_custom_dpi(self):
        """Test createFigure with custom DPI."""
        fig = createFigure(dpi=150)

        assert fig.get_figwidth() == 12
        assert fig.get_figheight() == 8
        assert fig.dpi == 150

        plt.close(fig)

    def test_createFigure_with_subplotAdj(self):
        """Test createFigure with subplot adjustments."""
        subplot_params = {
            "left": 0.1,
            "right": 0.9,
            "top": 0.9,
            "bottom": 0.1,
            "hspace": 0.2,
            "wspace": 0.2,
        }

        fig = createFigure(subplotAdj=subplot_params)

        # Check that figure was created successfully
        assert fig.get_figwidth() == 12
        assert fig.get_figheight() == 8

        plt.close(fig)

    def test_createFigure_kwargs(self):
        """Test createFigure with additional keyword arguments."""
        fig = createFigure(facecolor="lightblue", edgecolor="red")

        # Check that figure was created successfully
        assert fig.get_figwidth() == 12
        assert fig.get_figheight() == 8

        plt.close(fig)

    def test_createFigure_all_parameters(self):
        """Test createFigure with all parameters specified."""
        subplot_params = {"left": 0.2, "right": 0.8}

        fig = createFigure(
            figsize=(10, 5),
            dpi=200,
            subplotAdj=subplot_params,
            facecolor="white",
            edgecolor="black",
        )

        assert fig.get_figwidth() == 10
        assert fig.get_figheight() == 5
        assert fig.dpi == 200

        plt.close(fig)


class TestAddEqualAxes:
    """Test the add_equal_axes function."""

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_add_equal_axes_right(self):
        """Test adding axes to the right."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create some sample data
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        ax.plot(x, y, "b-", linewidth=2)
        ax.set_xlabel("X axis")
        ax.set_ylabel("Y axis")
        ax.set_title("Main Plot")
        ax.grid(True, alpha=0.3)

        # Add new axes to the right
        ax_new = add_equal_axes(ax, loc="right", pad=0.05, width=0.1)

        # Add content to new axes
        ax_new.plot(y, x, "r-", linewidth=2)
        ax_new.set_xlabel("Y values")
        ax_new.set_ylabel("X values")
        ax_new.grid(True, alpha=0.3)

        return fig

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_add_equal_axes_left(self):
        """Test adding axes to the left."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create some sample data
        x = np.linspace(0, 2 * np.pi, 100)
        y = np.cos(x)
        ax.plot(x, y, "g-", linewidth=2)
        ax.set_xlabel("X axis")
        ax.set_ylabel("Cosine values")
        ax.set_title("Main Plot")
        ax.grid(True, alpha=0.3)

        # Add new axes to the left
        ax_new = add_equal_axes(ax, loc="left", pad=0.05, width=0.15)

        # Add histogram to new axes
        ax_new.hist(y, bins=20, orientation="horizontal", alpha=0.7, color="orange")
        ax_new.set_xlabel("Frequency")
        ax_new.set_ylabel("Cosine values")

        return fig

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_add_equal_axes_top(self):
        """Test adding axes to the top."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create 2D data
        x = np.linspace(-2, 2, 50)
        y = np.linspace(-2, 2, 50)
        X, Y = np.meshgrid(x, y)
        Z = np.exp(-(X**2 + Y**2))

        # Main contour plot
        cs = ax.contourf(X, Y, Z, levels=20, cmap="viridis")
        ax.set_xlabel("X coordinate")
        ax.set_ylabel("Y coordinate")
        ax.set_title("2D Gaussian")

        # Add colorbar-like axes at top
        ax_new = add_equal_axes(ax, loc="top", pad=0.05, width=0.05)

        # Show mean values along x-axis
        mean_z = np.mean(Z, axis=0)
        ax_new.plot(x, mean_z, "k-", linewidth=2)
        ax_new.set_xlabel("X coordinate")
        ax_new.set_ylabel("Mean Z")
        ax_new.grid(True, alpha=0.3)

        return fig

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_add_equal_axes_bottom(self):
        """Test adding axes to the bottom."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create scatter plot
        np.random.seed(42)
        x = np.random.randn(200)
        y = 2 * x + np.random.randn(200) * 0.5

        ax.scatter(x, y, alpha=0.6, c="blue", s=30)
        ax.set_xlabel("X values")
        ax.set_ylabel("Y values")
        ax.set_title("Scatter Plot with Correlation")
        ax.grid(True, alpha=0.3)

        # Add marginal histogram at bottom
        ax_new = add_equal_axes(ax, loc="bottom", pad=0.05, width=0.1)

        ax_new.hist(x, bins=30, alpha=0.7, color="red", edgecolor="black")
        ax_new.set_xlabel("X values")
        ax_new.set_ylabel("Frequency")
        ax_new.grid(True, alpha=0.3)

        return fig

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_add_equal_axes_multiple(self):
        """Test adding multiple axes around a central plot."""
        fig, ax = plt.subplots(figsize=(10, 8))

        # Central heatmap
        np.random.seed(123)
        data = np.random.randn(20, 20)
        im = ax.imshow(data, cmap="RdBu_r", aspect="equal")
        ax.set_title("Central Heatmap")
        ax.set_xlabel("X index")
        ax.set_ylabel("Y index")

        # Add axes on all sides
        ax_right = add_equal_axes(ax, loc="right", pad=0.02, width=0.08)
        ax_top = add_equal_axes(ax, loc="top", pad=0.02, width=0.08)

        # Right: row means
        row_means = np.mean(data, axis=1)
        ax_right.plot(row_means, np.arange(len(row_means)), "k-", linewidth=2)
        ax_right.set_xlabel("Row mean")
        ax_right.set_ylim(ax.get_ylim())
        ax_right.grid(True, alpha=0.3)

        # Top: column means
        col_means = np.mean(data, axis=0)
        ax_top.plot(np.arange(len(col_means)), col_means, "k-", linewidth=2)
        ax_top.set_ylabel("Col mean")
        ax_top.set_xlim(ax.get_xlim())
        ax_top.grid(True, alpha=0.3)

        return fig

    def test_add_equal_axes_array_input(self):
        """Test add_equal_axes with array of axes as input."""
        fig, axes = plt.subplots(1, 2, figsize=(10, 4))

        # Plot on both axes
        x = np.linspace(0, 10, 100)
        axes[0].plot(x, np.sin(x))
        axes[1].plot(x, np.cos(x))

        # Add axes to the right of both subplots
        ax_new = add_equal_axes(axes, loc="right", pad=0.05, width=0.1)

        # Check that new axes was created
        assert ax_new is not None
        assert hasattr(ax_new, "plot")  # Should be a valid axes object

        plt.close(fig)

    def test_add_equal_axes_invalid_location(self):
        """Test add_equal_axes with invalid location parameter."""
        fig, ax = plt.subplots()

        with pytest.raises(ValueError):
            add_equal_axes(ax, loc="invalid_location", pad=0.05, width=0.1)

        plt.close(fig)

    def test_add_equal_axes_parameters(self):
        """Test add_equal_axes parameter validation."""
        fig, ax = plt.subplots()

        # Test with various parameter values
        ax_new1 = add_equal_axes(ax, loc="right", pad=0.1, width=0.2)
        ax_new2 = add_equal_axes(ax, loc="left", pad=0.01, width=0.05)

        # Check that axes were created successfully
        assert ax_new1 is not None
        assert ax_new2 is not None

        plt.close(fig)


class TestPlottingIntegration:
    """Integration tests for plotting module."""

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_plotting_with_climate_data(self, sample_2d_field):
        """Test plotting functions with realistic climate data."""
        fig, ax = plt.subplots(figsize=(10, 6))

        # Create a temperature map
        temp = sample_2d_field
        lat = temp.lat
        lon = temp.lon

        # Plot temperature field
        im = ax.contourf(lon, lat, temp, levels=20, cmap="RdYlBu_r")
        ax.set_xlabel("Longitude")
        ax.set_ylabel("Latitude")
        ax.set_title("Surface Temperature")

        # Add colorbar using add_equal_axes
        cax = add_equal_axes(ax, loc="right", pad=0.02, width=0.03)
        fig.colorbar(im, cax=cax, label="Temperature (K)")

        # Add marginal plots
        ax_top = add_equal_axes(ax, loc="top", pad=0.02, width=0.08)
        ax_right = add_equal_axes(
            ax, loc="right", pad=0.15, width=0.08
        )  # Further right

        # Zonal mean (average over longitude)
        zonal_mean = temp.mean(dim="lon")
        ax_right.plot(zonal_mean, lat, "k-", linewidth=2)
        ax_right.set_xlabel("Zonal mean (K)")
        ax_right.set_ylim(ax.get_ylim())
        ax_right.grid(True, alpha=0.3)

        # Meridional mean (average over latitude)
        meridional_mean = temp.mean(dim="lat")
        ax_top.plot(lon, meridional_mean, "k-", linewidth=2)
        ax_top.set_ylabel("Meridional mean (K)")
        ax_top.set_xlim(ax.get_xlim())
        ax_top.grid(True, alpha=0.3)

        return fig

    def test_plotting_error_handling(self):
        """Test error handling in plotting functions."""
        fig, ax = plt.subplots()

        # Test with invalid parameters
        with pytest.raises(ValueError):
            add_equal_axes(ax, loc="nowhere", pad=0.05, width=0.1)

        plt.close(fig)

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=20)
    def test_complex_layout(self, sample_climate_data):
        """Test complex multi-panel layout using add_equal_axes."""
        temp = sample_climate_data["temperature"]
        precip = sample_climate_data["precipitation"]

        fig, axes = plt.subplots(2, 2, figsize=(12, 10))
        fig.suptitle("Climate Data Analysis")

        # Temperature maps for different months
        temp_jan = temp.isel(time=0)  # January
        temp_jul = temp.isel(time=6)  # July

        im1 = axes[0, 0].contourf(
            temp.lon, temp.lat, temp_jan, levels=20, cmap="RdYlBu_r"
        )
        axes[0, 0].set_title("Temperature - January")
        axes[0, 0].set_xlabel("Longitude")
        axes[0, 0].set_ylabel("Latitude")

        im2 = axes[0, 1].contourf(
            temp.lon, temp.lat, temp_jul, levels=20, cmap="RdYlBu_r"
        )
        axes[0, 1].set_title("Temperature - July")
        axes[0, 1].set_xlabel("Longitude")

        # Precipitation maps
        precip_jan = precip.isel(time=0)
        precip_jul = precip.isel(time=6)

        im3 = axes[1, 0].contourf(
            precip.lon, precip.lat, precip_jan, levels=20, cmap="Blues"
        )
        axes[1, 0].set_title("Precipitation - January")
        axes[1, 0].set_xlabel("Longitude")
        axes[1, 0].set_ylabel("Latitude")

        im4 = axes[1, 1].contourf(
            precip.lon, precip.lat, precip_jul, levels=20, cmap="Blues"
        )
        axes[1, 1].set_title("Precipitation - July")
        axes[1, 1].set_xlabel("Longitude")

        # Add colorbars using add_equal_axes
        for i, (ax, im, label) in enumerate(
            [
                (axes[0, 0], im1, "Temperature (K)"),
                (axes[0, 1], im2, "Temperature (K)"),
                (axes[1, 0], im3, "Precipitation (mm/day)"),
                (axes[1, 1], im4, "Precipitation (mm/day)"),
            ]
        ):
            cax = add_equal_axes(ax, loc="right", pad=0.02, width=0.02)
            fig.colorbar(im, cax=cax, label=label if i % 2 == 0 else "")

        plt.tight_layout()
        return fig


if __name__ == "__main__":
    # Quick test runner
    pytest.main([__file__, "-v"])



"""Test the `gridfill` package."""

import os

import numpy as np
import numpy.ma as ma
import pytest

from skyborn.gridfill import fill


def _test_data_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


def _test_data_file(filename):
    return os.path.join(_test_data_dir(), filename)


def reference_solution(cyclic, initzonal):
    grid_data = ma.masked_invalid(np.load(_test_data_file("grid.npy")))
    cyclic_str = "_cyclic" if cyclic else ""
    zonal_str = "_zonal" if initzonal else ""
    soln_data = np.load(
        _test_data_file("soln{!s}{!s}.npy".format(cyclic_str, zonal_str))
    )
    return grid_data, soln_data


class GridFillTest(object):
    """Base class for `gridfill` tests."""

    # parameters to be set by each test subclass
    cyclic = None
    initzonal = None

    # parameters for relaxation scheme, same for all tests
    eps = 1e-4
    relax = 0.6
    itermax = 2000

    @classmethod
    def setup_class(cls):
        cls.grid, cls.soln = reference_solution(cls.cyclic, cls.initzonal)

    def test_single_grid(self):
        filled, c = fill(
            self.grid[0],
            1,
            0,
            self.eps,
            relax=self.relax,
            itermax=self.itermax,
            initzonal=self.initzonal,
            cyclic=self.cyclic,
            verbose=False,
        )
        self.assert_array_almost_equal(filled, self.soln[0])

    def test_multi_grid(self):
        filled, c = fill(
            self.grid,
            2,
            1,
            self.eps,
            relax=self.relax,
            itermax=self.itermax,
            initzonal=self.initzonal,
            cyclic=self.cyclic,
            verbose=False,
        )
        self.assert_array_almost_equal(filled, self.soln)

    def test_not_masked(self):
        with pytest.raises(TypeError):
            fill(
                self.grid.filled(fill_value=np.nan),
                self.eps,
                relax=self.relax,
                itermax=self.itermax,
                initzonal=self.initzonal,
                cyclic=self.cyclic,
                verbose=False,
            )

    def assert_array_almost_equal(self, a, b):
        np.testing.assert_array_almost_equal(a, b)


class TestFloat32(GridFillTest):
    """Test with 32-bit float input."""

    cyclic = False
    initzonal = False

    @classmethod
    def setup_class(cls):
        cls.grid, cls.soln = reference_solution(cls.cyclic, cls.initzonal)
        cls.grid = cls.grid.astype(np.float32)


class TestFillNonCyclicInitZero(GridFillTest):
    """Non-cyclic, initialized with zeros."""

    cyclic = False
    initzonal = False


class TestFillNonCyclicInitZonal(GridFillTest):
    """Non-cyclic, initialized with zonal mean."""

    cyclic = False
    initzonal = True


class TestFillCyclicInitZero(GridFillTest):
    """Cyclic, initialized with zeros."""

    cyclic = True
    initzonal = False


class TestFillCyclicInitZonal(GridFillTest):
    """Cyclic, initialized with zonal mean."""

    cyclic = True
    initzonal = True



"""
Troposphere calculation module.

This module provides WMO tropopause height calculation functionality
for atmospheric science research.

Main Functions:
    trop_wmo : WMO tropopause calculation for numpy arrays
    trop_wmo_profile : WMO tropopause calculation for 1D profiles

For xarray interface:
    >>> from skyborn.calc.troposphere.xarray import trop_wmo
"""

from . import xarray
from .tropopause import trop_wmo, trop_wmo_profile

__all__ = ["trop_wmo", "trop_wmo_profile"]



"""
Data format conversion utilities for the skyborn library.

This module provides functions for converting between different atmospheric
data formats, including GRIB to NetCDF conversion using eccodes.

Author: Qianye Su
Email: suqianye2000@gmail.com
"""

import subprocess  # For testing patches

from .grib_to_netcdf import (  # Private functions for testing
    GribToNetCDFError,
    _build_grib_to_netcdf_command,
    _check_grib_to_netcdf_available,
    _validate_grib_files,
    batch_convert_grib_to_nc,
    convert_grib_to_nc,
    convert_grib_to_nc_simple,
    grib2nc,
    grib_to_netcdf,
)



"""Tests for error handling in `windspharm`."""

from __future__ import absolute_import

import numpy as np
import numpy.ma as ma
import pytest

from skyborn.windspharm.tests import VectorWindTest, solvers

from .reference import reference_solutions


class ErrorHandlersTest(VectorWindTest):
    """Base class for all error handler tests."""

    interface = None
    gridtype = None

    @classmethod
    def setup_class(cls):
        msg = "missing dependencies required to test the {!s} interface"
        if cls.interface not in solvers:
            pytest.skip(msg.format(cls.interface))


# ----------------------------------------------------------------------------
# Tests for the standard interface


class TestStandardErrorHandlers(ErrorHandlersTest):
    """Standard interface error handler tests."""

    interface = "standard"
    gridtype = "regular"

    def test_masked_values(self):
        # masked values in inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        mask = np.empty(solution["uwnd"].shape, dtype=bool)
        mask[:] = False
        mask[1, 1] = True
        u = ma.array(solution["uwnd"], mask=mask, fill_value=1.0e20)
        v = ma.array(solution["vwnd"], mask=mask, fill_value=1.0e20)
        with pytest.raises(ValueError):
            solvers[self.interface](u, v, gridtype=self.gridtype)

    def test_nan_values(self):
        # NaN values in inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        solution["vwnd"][1, 1] = np.nan
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"], solution["vwnd"], gridtype=self.gridtype
            )

    def test_invalid_shape_components(self):
        # invalid shape inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"][np.newaxis].repeat(2, axis=0),
                solution["vwnd"][np.newaxis].repeat(2, axis=0),
                gridtype=self.gridtype,
            )

    def test_different_shape_components(self):
        # different shape inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"], solution["vwnd"][:-1], gridtype=self.gridtype
            )

    def test_invalid_rank_components(self):
        # invalid rank inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"][..., np.newaxis, np.newaxis],
                solution["vwnd"][..., np.newaxis, np.newaxis],
                gridtype=self.gridtype,
            )

    def test_different_rank_components(self):
        # different rank inputs should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"][..., np.newaxis],
                solution["vwnd"],
                gridtype=self.gridtype,
            )

    def test_invalid_gridtype(self):
        # invalid grid type specification should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        with pytest.raises(ValueError):
            solvers[self.interface](
                solution["uwnd"], solution["vwnd"], gridtype="curvilinear"
            )

    def test_gradient_masked_values(self):
        # masked values in gradient input should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](
            solution["uwnd"], solution["vwnd"], gridtype=self.gridtype
        )
        mask = np.empty(solution["uwnd"].shape, dtype=np.bool_)
        mask[:] = False
        mask[1, 1] = True
        chi = ma.array(solution["chi"], mask=mask, fill_value=1.0e20)
        with pytest.raises(ValueError):
            vw.gradient(chi)

    def test_gradient_nan_values(self):
        # NaN values in gradient input should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](
            solution["uwnd"], solution["vwnd"], gridtype=self.gridtype
        )
        solution["chi"][1, 1] = np.nan
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"])

    def test_gradient_invalid_shape(self):
        # input to gradient of different shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](
            solution["uwnd"], solution["vwnd"], gridtype=self.gridtype
        )
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"][:-1])


# ----------------------------------------------------------------------------
# Tests for the iris interface


class TestIrisErrorHandlers(ErrorHandlersTest):
    """Iris interface error handler tests."""

    interface = "iris"
    gridtype = "regular"

    def test_non_cube_input(self):
        # input not an iris cube should raise an error
        solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_different_shape_components(self):
        # inputs not the same shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        solution["vwnd"].transpose([1, 0])
        with pytest.raises(ValueError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_unknown_grid(self):
        # inputs where a lat-lon grid cannot be identified should raise an
        # error
        solution = reference_solutions(self.interface, self.gridtype)
        solution["vwnd"].coord("latitude").rename("unknown")
        with pytest.raises(ValueError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_gradient_non_cube_input(self):
        # input to gradient not an iris cube should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        dummy_solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            vw.gradient(dummy_solution["chi"])

    def test_gradient_different_shape(self):
        # input to gradient of different shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"][:-1])

    def test_gradient_unknown_grid(self):
        # input to gradient with no identifiable grid should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        solution["chi"].coord("latitude").rename("unknown")
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"])

    def test_truncate_non_cube_input(self):
        # input to truncate not an iris cube should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        dummy_solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            vw.truncate(dummy_solution["chi"])

    def test_truncate_different_shape(self):
        # input to truncate of different shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        with pytest.raises(ValueError):
            vw.truncate(solution["chi"][:-1])

    def test_truncate_unknown_grid(self):
        # input to truncate with no identifiable grid should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        solution["chi"].coord("latitude").rename("unknown")
        with pytest.raises(ValueError):
            vw.truncate(solution["chi"])


# ----------------------------------------------------------------------------
# Tests for the xarray interface


class TestXarrayErrorHandlers(ErrorHandlersTest):
    """xarray interface error handler tests."""

    interface = "xarray"
    gridtype = "regular"

    def test_non_dataarray_input(self):
        # input not an xarray.DataArray should raise an error
        solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_different_shape_components(self):
        # inputs not the same shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        solution["vwnd"] = solution["vwnd"].transpose()
        with pytest.raises(ValueError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_unknown_grid(self):
        # inputs where a lat-lon grid cannot be identified should raise an
        # error
        solution = reference_solutions(self.interface, self.gridtype)
        solution["vwnd"].coords.update(
            {"unknown": ("latitude", solution["vwnd"].coords["latitude"].values)}
        )
        solution["vwnd"] = solution["vwnd"].swap_dims({"latitude": "unknown"})
        del solution["vwnd"].coords["latitude"]
        with pytest.raises(ValueError):
            solvers[self.interface](solution["uwnd"], solution["vwnd"])

    def test_gradient_non_dataarray_input(self):
        # input to gradient not an xarray.DataArray should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        dummy_solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            vw.gradient(dummy_solution["chi"])

    def test_gradient_different_shape(self):
        # input to gradient of different shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"][:-1])

    def test_gradient_unknown_grid(self):
        # input to gradient with no identifiable grid should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        solution["chi"].coords.update(
            {"unknown": ("latitude", solution["chi"].coords["latitude"].values)}
        )
        solution["chi"] = solution["chi"].swap_dims({"latitude": "unknown"})
        del solution["chi"].coords["latitude"]
        with pytest.raises(ValueError):
            vw.gradient(solution["chi"])

    def test_truncate_non_dataarray_input(self):
        # input to truncate not an xarray.DataArray should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        dummy_solution = reference_solutions("standard", self.gridtype)
        with pytest.raises(TypeError):
            vw.truncate(dummy_solution["chi"])

    def test_truncate_different_shape(self):
        # input to truncate of different shape should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        with pytest.raises(ValueError):
            vw.truncate(solution["chi"][:-1])

    def test_truncate_unknown_grid(self):
        # input to truncate with no identifiable grid should raise an error
        solution = reference_solutions(self.interface, self.gridtype)
        vw = solvers[self.interface](solution["uwnd"], solution["vwnd"])
        solution["chi"].coords.update(
            {"unknown": ("latitude", solution["chi"].coords["latitude"].values)}
        )
        solution["chi"] = solution["chi"].swap_dims({"latitude": "unknown"})
        del solution["chi"].coords["latitude"]
        with pytest.raises(ValueError):
            vw.truncate(solution["chi"])



"""
Tests for skyborn.plot.curved_quiver_plot module.

This module tests the curved quiver plotting functionality,
including the curved_quiver function and CurvedQuiverLegend class.
"""

from unittest.mock import MagicMock, Mock, patch

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr
from matplotlib.patches import FancyArrowPatch
from matplotlib.text import Text

from skyborn.plot.curved_quiver_plot import (
    CurvedQuiverLegend,
    add_curved_quiverkey,
    curved_quiver,
)
from skyborn.plot.modplot import CurvedQuiverplotSet


class TestCurvedQuiver:
    """Test the curved_quiver function."""

    @pytest.fixture
    def sample_data(self):
        """Create sample wind data for testing."""
        # Create 2D coordinate arrays
        x = np.linspace(-10, 10, 15)
        y = np.linspace(-5, 5, 10)

        # Create 2D grid
        X, Y = np.meshgrid(x, y)

        # Create sample wind field (circular pattern)
        u = -Y * 0.5  # u component
        v = X * 0.3  # v component

        # Create xarray Dataset
        ds = xr.Dataset(
            {
                "u": (["y", "x"], u),
                "v": (["y", "x"], v),
            },
            coords={"x": (["x"], x), "y": (["y"], y)},
        )

        return ds

    def test_curved_quiver_basic(self, sample_data):
        """Test basic curved_quiver functionality."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create curved quiver plot
        result = curved_quiver(
            sample_data, x="x", y="y", u="u", v="v", ax=ax, density=1
        )

        # Check return type
        assert isinstance(result, CurvedQuiverplotSet)

        # Check that result has required attributes
        assert hasattr(result, "lines")
        assert hasattr(result, "arrows")
        assert hasattr(result, "resolution")
        assert hasattr(result, "magnitude")

        plt.close(fig)

    def test_curved_quiver_with_parameters(self, sample_data):
        """Test curved_quiver with various parameters."""
        fig, ax = plt.subplots(figsize=(8, 6))

        result = curved_quiver(
            sample_data,
            x="x",
            y="y",
            u="u",
            v="v",
            ax=ax,
            density=2,
            linewidth=2.0,
            color="red",
            arrowsize=1.5,
            arrowstyle="->",
        )

        assert isinstance(result, CurvedQuiverplotSet)
        assert result.linewidth == 2.0
        assert result.color == "red"
        assert result.arrowsize == 1.5
        assert result.arrowstyle == "->"

        plt.close(fig)

    def test_curved_quiver_no_axes(self, sample_data):
        """Test curved_quiver without providing axes."""
        # This should use current axes
        plt.figure(figsize=(8, 6))

        result = curved_quiver(sample_data, x="x", y="y", u="u", v="v", density=1)

        assert isinstance(result, CurvedQuiverplotSet)
        plt.close()

    def test_curved_quiver_integration_directions(self, sample_data):
        """Test different integration directions."""
        directions = ["forward", "backward", "both"]

        for direction in directions:
            fig, ax = plt.subplots(figsize=(6, 4))

            result = curved_quiver(
                sample_data,
                x="x",
                y="y",
                u="u",
                v="v",
                ax=ax,
                integration_direction=direction,
            )

            assert isinstance(result, CurvedQuiverplotSet)
            assert result.integration_direction == direction

            plt.close(fig)

    def test_curved_quiver_with_start_points(self, sample_data):
        """Test curved_quiver with custom start points."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Define custom start points
        start_points = np.array([[-5, -2], [0, 0], [5, 2]])

        result = curved_quiver(
            sample_data, x="x", y="y", u="u", v="v", ax=ax, start_points=start_points
        )

        assert isinstance(result, CurvedQuiverplotSet)
        np.testing.assert_array_equal(result.start_points, start_points)

        plt.close(fig)

    def test_curved_quiver_with_transform(self, sample_data):
        """Test curved_quiver with transform parameter."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Skip this test due to complex mock transform requirements
        plt.close(fig)
        pytest.skip("Transform test requires complex matplotlib transform mocking")

    def test_curved_quiver_with_platecarree_transform(self, sample_data):
        """Test curved_quiver with PlateCarree transform."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Skip this test due to complex mock transform requirements
        plt.close(fig)
        pytest.skip(
            "PlateCarree transform test requires complex matplotlib transform mocking"
        )

    def test_curved_quiver_with_colormap(self, sample_data):
        """Test curved_quiver with colormap settings."""
        fig, ax = plt.subplots(figsize=(8, 6))

        # Create color data
        color_data = np.random.randn(10, 15)

        result = curved_quiver(
            sample_data,
            x="x",
            y="y",
            u="u",
            v="v",
            ax=ax,
            color=color_data,
            cmap="viridis",
            norm=plt.Normalize(vmin=0, vmax=1),
        )

        assert isinstance(result, CurvedQuiverplotSet)
        plt.close(fig)

    def test_curved_quiver_with_zorder(self, sample_data):
        """Test curved_quiver with zorder parameter."""
        fig, ax = plt.subplots(figsize=(8, 6))

        result = curved_quiver(
            sample_data, x="x", y="y", u="u", v="v", ax=ax, zorder=10
        )

        assert isinstance(result, CurvedQuiverplotSet)
        plt.close(fig)

    def test_curved_quiver_grains_parameter(self, sample_data):
        """Test curved_quiver with different grains settings."""
        fig, ax = plt.subplots(figsize=(8, 6))

        result = curved_quiver(
            sample_data,
            x="x",
            y="y",
            u="u",
            v="v",
            ax=ax,
            grains=20,
            broken_streamlines=False,
        )

        assert isinstance(result, CurvedQuiverplotSet)
        plt.close(fig)

    @patch("skyborn.plot.modplot.velovect")
    def test_curved_quiver_calls_velovect(self, mock_velovect, sample_data):
        """Test that curved_quiver properly calls velovect."""
        mock_result = Mock(spec=CurvedQuiverplotSet)
        mock_velovect.return_value = mock_result

        fig, ax = plt.subplots(figsize=(8, 6))

        result = curved_quiver(
            sample_data, x="x", y="y", u="u", v="v", ax=ax, density=2, linewidth=1.5
        )

        # Check that velovect was called
        mock_velovect.assert_called_once()

        # Check some of the arguments passed to velovect
        call_args = mock_velovect.call_args
        assert call_args[0][0] == ax  # First argument should be axes
        assert call_args[1]["density"] == 2
        assert call_args[1]["linewidth"] == 1.5

        assert result == mock_result
        plt.close(fig)


class TestCurvedQuiverLegend:
    """Test the CurvedQuiverLegend class."""

    @pytest.fixture
    def mock_curved_quiver_set(self):
        """Create a mock CurvedQuiverplotSet for testing."""
        mock_set = Mock(spec=CurvedQuiverplotSet)
        mock_set.resolution = 0.5
        mock_set.magnitude = np.array([[0.5, 1.0], [1.5, 2.0]])
        return mock_set

    @pytest.fixture
    def sample_axes(self):
        """Create sample matplotlib axes."""
        fig, ax = plt.subplots(figsize=(8, 6))
        return ax, fig

    def test_curved_quiver_legend_basic(self, sample_axes, mock_curved_quiver_set):
        """Test basic CurvedQuiverLegend creation."""
        ax, fig = sample_axes

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_curved_quiver_set, U=2.0, units="m/s"
        )

        # Check basic attributes
        assert legend.ax == ax
        assert legend.curved_quiver_set == mock_curved_quiver_set
        assert legend.U == 2.0
        assert legend.units == "m/s"
        assert legend.show_units == True

        # Check that components were created
        assert hasattr(legend, "patch")
        assert hasattr(legend, "arrow")
        assert hasattr(legend, "text")

        plt.close(fig)

    def test_curved_quiver_legend_different_locations(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test CurvedQuiverLegend with different locations."""
        ax, fig = sample_axes

        locations = ["lower left", "lower right", "upper left", "upper right"]

        for loc in locations:
            legend = CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=1.5,
                units="m/s",
                loc=loc,
            )

            assert legend.loc == loc
            assert hasattr(legend, "patch")
            assert hasattr(legend, "arrow")
            assert hasattr(legend, "text")

        plt.close(fig)

    def test_curved_quiver_legend_label_positions(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test CurvedQuiverLegend with different label positions."""
        ax, fig = sample_axes

        label_positions = ["N", "S", "E", "W"]

        for labelpos in label_positions:
            legend = CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=2.0,
                units="km/h",
                labelpos=labelpos,
            )

            assert legend.labelpos == labelpos
            assert hasattr(legend, "text")

        plt.close(fig)

    def test_curved_quiver_legend_custom_properties(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test CurvedQuiverLegend with custom properties."""
        ax, fig = sample_axes

        arrow_props = {"color": "red", "linewidth": 2}
        patch_props = {"facecolor": "lightblue", "alpha": 0.8}
        text_props = {"fontsize": 12, "color": "darkblue"}

        legend = CurvedQuiverLegend(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=3.0,
            units="kt",
            width=0.2,
            height=0.1,
            max_arrow_length=0.06,
            arrow_props=arrow_props,
            patch_props=patch_props,
            text_props=text_props,
            padding=0.02,
            margin=0.03,
        )

        assert legend.width == 0.2
        assert legend.height == 0.1
        assert legend.max_arrow_length == 0.06
        assert legend.arrow_props == arrow_props
        assert legend.patch_props == patch_props
        assert legend.text_props == text_props
        assert legend.padding == 0.02
        assert legend.margin == 0.03

        plt.close(fig)

    def test_curved_quiver_legend_positions(self, sample_axes, mock_curved_quiver_set):
        """Test different legend positions."""
        ax, fig = sample_axes
        positions = ["lower left", "lower right", "upper left", "upper right"]

        for pos in positions:
            legend = CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=2.0,
                units="m/s",
                loc=pos,
            )

            assert legend.loc == pos
            # Position should be calculated
            assert hasattr(legend, "x")
            assert hasattr(legend, "y")

        plt.close(fig)

    def test_curved_quiver_legend_label_positions(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test different label positions."""
        ax, fig = sample_axes
        label_positions = ["N", "S", "E", "W"]

        for labelpos in label_positions:
            legend = CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=2.0,
                units="m/s",
                labelpos=labelpos,
            )

            assert legend.labelpos == labelpos

        plt.close(fig)

    def test_curved_quiver_legend_no_units(self, sample_axes, mock_curved_quiver_set):
        """Test legend with no units (auto-centering)."""
        ax, fig = sample_axes

        legend = CurvedQuiverLegend(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=5.0,
            units="",  # Empty units
        )

        assert legend.units == ""
        assert legend.show_units == False
        assert legend.center_label == True  # Should auto-center
        assert legend.text_content == "5.0"  # Should not include units

        plt.close(fig)

    def test_curved_quiver_legend_with_units(self, sample_axes, mock_curved_quiver_set):
        """Test legend with units."""
        ax, fig = sample_axes

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_curved_quiver_set, U=3.5, units="m/s"
        )

        assert legend.units == "m/s"
        assert legend.show_units == True
        assert legend.center_label == False

        plt.close(fig)

    def test_curved_quiver_legend_custom_properties(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test legend with custom properties."""
        ax, fig = sample_axes

        arrow_props = {"color": "blue", "linewidth": 2}
        patch_props = {"facecolor": "yellow", "alpha": 0.8}
        text_props = {"fontsize": 12, "color": "red"}

        legend = CurvedQuiverLegend(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=2.0,
            units="m/s",
            arrow_props=arrow_props,
            patch_props=patch_props,
            text_props=text_props,
        )

        # Check that properties were applied
        assert legend.arrow_props["color"] == "blue"
        assert legend.arrow_props["linewidth"] == 2
        assert legend.patch_props["facecolor"] == "yellow"
        assert legend.patch_props["alpha"] == 0.8
        assert legend.text_props["fontsize"] == 12
        assert legend.text_props["color"] == "red"

        plt.close(fig)

    def test_curved_quiver_legend_arrow_length_calculation(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test arrow length calculation."""
        ax, fig = sample_axes

        # Test with different wind speeds
        wind_speeds = [1.0, 2.0, 5.0, 10.0]

        for U in wind_speeds:
            legend = CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=U,
                reference_speed=2.0,
                max_arrow_length=0.08,
            )

            # Arrow length should be proportional to wind speed
            expected_scale = U / 2.0  # reference_speed = 2.0
            assert hasattr(legend, "arrow_length")
            assert legend.arrow_length > 0

        plt.close(fig)

    def test_curved_quiver_legend_size_adjustment(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test that legend box size adjusts for content."""
        ax, fig = sample_axes

        # Create legend with long text
        legend = CurvedQuiverLegend(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=15.5,
            units="very_long_unit_string",
            width=0.1,  # Initial small width
            height=0.05,  # Initial small height
        )

        # Box should have been resized to fit content
        assert legend.width >= 0.1  # Should be at least the initial width
        assert legend.height >= 0.05  # Should be at least the initial height

        plt.close(fig)

    def test_curved_quiver_legend_invalid_position(
        self, sample_axes, mock_curved_quiver_set
    ):
        """Test invalid legend position raises error."""
        ax, fig = sample_axes

        with pytest.raises(ValueError, match="loc must be one of"):
            CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_curved_quiver_set,
                U=2.0,
                units="m/s",
                loc="invalid_position",
            )

        plt.close(fig)


class TestAddCurvedQuiverkey:
    """Test the add_curved_quiverkey convenience function."""

    @pytest.fixture
    def mock_curved_quiver_set(self):
        """Create a mock CurvedQuiverplotSet for testing."""
        mock_set = Mock(spec=CurvedQuiverplotSet)
        mock_set.resolution = 0.5
        mock_set.magnitude = np.array([[0.5, 1.0], [1.5, 2.0]])
        return mock_set

    def test_add_curved_quiverkey_basic(self, mock_curved_quiver_set):
        """Test basic add_curved_quiverkey functionality."""
        fig, ax = plt.subplots(figsize=(8, 6))

        result = add_curved_quiverkey(
            ax=ax, curved_quiver_set=mock_curved_quiver_set, U=2.0, units="m/s"
        )

        # Should return a CurvedQuiverLegend instance
        assert isinstance(result, CurvedQuiverLegend)
        assert result.U == 2.0
        assert result.units == "m/s"
        assert result.loc == "lower right"  # default
        assert result.labelpos == "E"  # default

        plt.close(fig)

    def test_add_curved_quiverkey_with_parameters(self, mock_curved_quiver_set):
        """Test add_curved_quiverkey with custom parameters."""
        fig, ax = plt.subplots(figsize=(8, 6))

        result = add_curved_quiverkey(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=5.0,
            units="knots",
            loc="upper left",
            labelpos="W",
            width=0.2,
            height=0.1,
        )

        assert isinstance(result, CurvedQuiverLegend)
        assert result.U == 5.0
        assert result.units == "knots"
        assert result.loc == "upper left"
        assert result.labelpos == "W"
        # Width might be adjusted during initialization, so just check it's reasonable
        assert 0.1 < result.width < 0.5
        assert result.height == 0.1

        plt.close(fig)

    def test_add_curved_quiverkey_kwargs_passed(self, mock_curved_quiver_set):
        """Test that additional kwargs are passed to CurvedQuiverLegend."""
        fig, ax = plt.subplots(figsize=(8, 6))

        custom_arrow_props = {"color": "green", "linewidth": 3}

        result = add_curved_quiverkey(
            ax=ax,
            curved_quiver_set=mock_curved_quiver_set,
            U=3.0,
            arrow_props=custom_arrow_props,
            margin=0.05,
        )

        assert isinstance(result, CurvedQuiverLegend)
        assert result.arrow_props["color"] == "green"
        assert result.arrow_props["linewidth"] == 3
        assert result.margin == 0.05

        plt.close(fig)


class TestCurvedQuiverIntegration:
    """Integration tests for curved quiver plotting."""

    @pytest.fixture
    def wind_data(self):
        """Create realistic wind data for integration testing."""
        # Create coordinate arrays
        lon = np.linspace(-10, 10, 20)
        lat = np.linspace(-5, 5, 15)

        # Create 2D grid
        LON, LAT = np.meshgrid(lon, lat)

        # Create realistic wind pattern (westerly jet with some curvature)
        u = 10 * (1 - (LAT / 5) ** 2) * np.exp(-0.1 * LON**2)  # Westerly component
        v = 2 * np.sin(LON * np.pi / 10) * np.exp(-0.1 * LAT**2)  # Meridional component

        # Add some noise
        np.random.seed(42)
        u += np.random.normal(0, 0.5, u.shape)
        v += np.random.normal(0, 0.3, v.shape)

        # Create xarray Dataset
        ds = xr.Dataset(
            {
                "u": (["lat", "lon"], u),
                "v": (["lat", "lon"], v),
            },
            coords={"lon": (["lon"], lon), "lat": (["lat"], lat)},
        )

        return ds

    def test_curved_quiver_with_legend_integration(self, wind_data):
        """Test full integration of curved quiver with legend."""
        fig, ax = plt.subplots(figsize=(10, 8))

        # Create curved quiver plot
        quiver_set = curved_quiver(
            wind_data,
            x="lon",
            y="lat",
            u="u",
            v="v",
            ax=ax,
            density=1.5,
            color="blue",
            linewidth=1.5,
            arrowsize=1.2,
        )

        # Add legend
        legend = add_curved_quiverkey(
            ax=ax,
            curved_quiver_set=quiver_set,
            U=10.0,
            units="m/s",
            loc="upper right",
            labelpos="E",
        )

        # Verify everything was created properly
        assert isinstance(quiver_set, CurvedQuiverplotSet)
        assert isinstance(legend, CurvedQuiverLegend)

        # Check that axes has the expected artists
        # Lines and patches should be added to the axes
        assert len(ax.collections) > 0  # Should have LineCollection
        # Should have arrow patches and legend patch
        assert len(ax.patches) > 0

        # Set some basic plot properties
        ax.set_xlabel("Longitude")
        ax.set_ylabel("Latitude")
        ax.set_title("Curved Quiver Plot with Legend")
        ax.grid(True, alpha=0.3)

        plt.close(fig)

    def test_multiple_legends_different_positions(self, wind_data):
        """Test multiple legends in different positions."""
        fig, ax = plt.subplots(figsize=(12, 8))

        # Create curved quiver plot
        quiver_set = curved_quiver(
            wind_data, x="lon", y="lat", u="u", v="v", ax=ax, density=1
        )

        # Add multiple legends
        positions = ["lower left", "lower right", "upper left"]
        speeds = [5.0, 10.0, 15.0]

        legends = []
        for pos, speed in zip(positions, speeds):
            legend = add_curved_quiverkey(
                ax=ax,
                curved_quiver_set=quiver_set,
                U=speed,
                units="m/s",
                loc=pos,
                width=0.15,
                height=0.08,
            )
            legends.append(legend)

        # All legends should be created
        assert len(legends) == 3
        for legend in legends:
            assert isinstance(legend, CurvedQuiverLegend)

        plt.close(fig)


class TestCurvedQuiverLegendUntestedMethods:
    """Test previously untested methods in CurvedQuiverLegend."""

    @pytest.fixture
    def sample_axes_and_set(self):
        """Create sample axes and mock quiver set."""
        fig, ax = plt.subplots(figsize=(8, 6))
        mock_set = Mock(spec=CurvedQuiverplotSet)
        mock_set.resolution = 0.5
        mock_set.magnitude = np.array([[0.5, 1.0], [1.5, 2.0]])
        return ax, fig, mock_set

    def test_create_arrow_method(self, sample_axes_and_set):
        """Test _create_arrow method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test arrow creation
        arrow = legend._create_arrow(0.1, 0.1, 0.2, 0.2)
        assert isinstance(arrow, FancyArrowPatch)

        plt.close(fig)

    def test_create_text_method(self, sample_axes_and_set):
        """Test _create_text method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test text creation
        text = legend._create_text(0.5, 0.5, h_align="left", v_align="top")
        assert isinstance(text, Text)

        plt.close(fig)

    def test_calculate_box_center(self, sample_axes_and_set):
        """Test _calculate_box_center method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test box center calculation
        center_x, center_y = legend._calculate_box_center()
        assert isinstance(center_x, (int, float))
        assert isinstance(center_y, (int, float))

        plt.close(fig)

    def test_calculate_horizontal_arrow_positions(self, sample_axes_and_set):
        """Test _calculate_horizontal_arrow_positions method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test horizontal arrow position calculation
        center_x, center_y = legend._calculate_box_center()
        start_x, end_x = legend._calculate_horizontal_arrow_positions(
            center_x, center_y, False, True  # label_left=True
        )

        assert isinstance(start_x, (int, float))
        assert isinstance(end_x, (int, float))
        assert start_x < end_x  # start should be left of end

        plt.close(fig)

    def test_calculate_vertical_arrow_positions(self, sample_axes_and_set):
        """Test _calculate_vertical_arrow_positions method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test vertical arrow position calculation
        center_x = 0.5
        start_x, end_x = legend._calculate_vertical_arrow_positions(center_x, False)

        assert isinstance(start_x, (int, float))
        assert isinstance(end_x, (int, float))
        assert start_x < end_x  # start should be left of end

        plt.close(fig)

    def test_calculate_text_position_horizontal(self, sample_axes_and_set):
        """Test _calculate_text_position_horizontal method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test horizontal text position calculation
        for labelpos in ["N", "S", "E", "W"]:
            legend.labelpos = labelpos
            text_x, text_y, h_align, v_align = (
                legend._calculate_text_position_horizontal(
                    0.1, 0.1, 0.2, 0.2, 0.05, 0.02
                )
            )

            assert isinstance(text_x, (int, float))
            assert isinstance(text_y, (int, float))
            assert h_align in ["left", "right", "center"]
            assert v_align in ["top", "bottom", "center"]

        plt.close(fig)

    def test_calculate_text_position_vertical(self, sample_axes_and_set):
        """Test _calculate_text_position_vertical method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test vertical text position calculation - only test valid cases
        for labelpos in ["N", "S"]:
            legend.labelpos = labelpos
            text_x, text_y, h_align, v_align = legend._calculate_text_position_vertical(
                0.1, 0.1, 0.02, False, labelpos == "N"  # label_above for N
            )

            assert isinstance(text_x, (int, float))
            assert isinstance(text_y, (int, float))
            assert h_align in ["left", "right", "center"]
            assert v_align in ["top", "bottom", "center"]

        plt.close(fig)

    def test_position_arrow_and_text(self, sample_axes_and_set):
        """Test _position_arrow_and_text method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test arrow and text positioning
        legend._position_arrow_and_text(0.1, 0.05)  # text_width, text_height

        # Should have created arrow and text objects
        assert hasattr(legend, "arrow")
        assert hasattr(legend, "text")

        plt.close(fig)

    def test_calculate_arrow_length(self, sample_axes_and_set):
        """Test _calculate_arrow_length method."""
        ax, fig, mock_set = sample_axes_and_set

        # Test with different reference speeds
        for ref_speed in [1.0, 5.0, 10.0]:
            legend = CurvedQuiverLegend(
                ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
            )
            # Set reference speed after initialization if attribute exists
            if hasattr(legend, "reference_speed"):
                legend.reference_speed = ref_speed

            arrow_length = legend._calculate_arrow_length()
            assert arrow_length > 0
            # Don't assert max length constraint since it depends on implementation
            assert isinstance(arrow_length, (int, float))

        plt.close(fig)

    def test_calculate_position(self, sample_axes_and_set):
        """Test _calculate_position method."""
        ax, fig, mock_set = sample_axes_and_set

        # Test with different locations
        for loc in ["lower left", "lower right", "upper left", "upper right"]:
            legend = CurvedQuiverLegend(
                ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s", loc=loc
            )

            legend._calculate_position()

            # Should set x and y coordinates
            assert hasattr(legend, "x")
            assert hasattr(legend, "y")
            assert 0 <= legend.x <= 1
            assert 0 <= legend.y <= 1

        plt.close(fig)

    def test_setup_props_methods(self, sample_axes_and_set):
        """Test property setup methods."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test _setup_props method
        defaults = {"color": "black", "linewidth": 1}
        user_props = {"color": "red"}
        result = legend._setup_props(user_props, defaults)

        assert result["color"] == "red"
        assert result["linewidth"] == 1

        # Test individual setup methods
        arrow_props = legend._setup_arrow_props({"color": "blue"})
        assert arrow_props["color"] == "blue"

        patch_props = legend._setup_patch_props({"facecolor": "yellow"})
        assert patch_props["facecolor"] == "yellow"

        text_props = legend._setup_text_props({"fontsize": 14})
        assert text_props["fontsize"] == 14

        plt.close(fig)

    def test_set_figure_method(self, sample_axes_and_set):
        """Test set_figure method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Test that legend has figure reference
        assert legend.figure is not None
        assert legend.figure == fig

        # Test that trying to change figure would raise an error
        new_fig = plt.figure(figsize=(10, 8))
        try:
            legend.set_figure(new_fig)
            # If it doesn't raise, that's also valid behavior
            assert True
        except RuntimeError:
            # Expected behavior - can't change figure
            assert True
        finally:
            plt.close(new_fig)

        plt.close(fig)

    def test_draw_method(self, sample_axes_and_set):
        """Test draw method."""
        ax, fig, mock_set = sample_axes_and_set

        legend = CurvedQuiverLegend(
            ax=ax, curved_quiver_set=mock_set, U=2.0, units="m/s"
        )

        # Mock renderer
        from matplotlib.backend_bases import RendererBase

        mock_renderer = Mock(spec=RendererBase)

        # Test draw method (should not raise exceptions)
        try:
            legend.draw(mock_renderer)
        except Exception as e:
            # Some drawing operations might fail in test environment
            # but method should exist and be callable
            assert callable(legend.draw)

        plt.close(fig)


class TestCurvedQuiverErrorHandling:
    """Test error handling in curved quiver plotting."""

    def test_curved_quiver_with_invalid_data(self):
        """Test curved_quiver with invalid data shapes."""
        # Create compatible data with same shape for both u and v
        u_data = np.random.randn(10, 15)
        v_data = np.random.randn(10, 15)  # Same shape as u

        x = np.linspace(-10, 10, 15)
        y = np.linspace(-5, 5, 10)

        # Test with missing variables should raise error
        ds = xr.Dataset(
            {"u": (["y", "x"], u_data)}, coords={"x": x, "y": y}  # Missing v variable
        )

        fig, ax = plt.subplots()

        # Should handle gracefully or raise appropriate error
        with pytest.raises((ValueError, KeyError)):
            curved_quiver(ds, x="x", y="y", u="u", v="v", ax=ax)

        plt.close(fig)

    def test_legend_with_invalid_location(self):
        """Test legend creation with invalid location."""
        fig, ax = plt.subplots()
        mock_set = Mock(spec=CurvedQuiverplotSet)
        mock_set.resolution = 0.5
        mock_set.magnitude = np.array([[1.0, 2.0]])

        # Should raise error for invalid location
        with pytest.raises(ValueError):
            CurvedQuiverLegend(
                ax=ax,
                curved_quiver_set=mock_set,
                U=2.0,
                units="m/s",
                loc="invalid_location",
            )

        plt.close(fig)



"""
Setup script for Skyborn - Mixed build system with meson for Fortran modules
"""

import os
import shutil
import subprocess
import sys
from pathlib import Path

import numpy as np
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop
from setuptools.command.install import install

# Check if Cython is available
try:
    from Cython.Build import cythonize

    HAVE_CYTHON = True
except ImportError:
    HAVE_CYTHON = False

# Check if we're in documentation build mode
DOCS_BUILD_MODE = (
    os.environ.get("SKYBORN_DOCS_BUILD") == "1" or os.environ.get("SKIP_FORTRAN") == "1"
)

if DOCS_BUILD_MODE:
    print("📚 Documentation build mode detected - skipping Fortran compilation")
else:
    # Force gfortran compiler usage
    os.environ["FC"] = os.environ.get("FC", "gfortran")
    os.environ["F77"] = os.environ.get("F77", "gfortran")
    os.environ["F90"] = os.environ.get("F90", "gfortran")
    os.environ["CC"] = os.environ.get("CC", "gcc")


# Check if gfortran is available
def check_gfortran():
    """Check if gfortran is available"""
    try:
        result = subprocess.run(
            ["gfortran", "--version"], capture_output=True, text=True, timeout=10
        )
        if result.returncode == 0:
            print(
                f"Found gfortran: {result.stdout.split()[4] if len(result.stdout.split()) > 4 else 'unknown version'}"
            )
            return True
    except (subprocess.TimeoutExpired, subprocess.SubprocessError, FileNotFoundError):
        pass

    print("Warning: gfortran not found. Fortran extensions may not build correctly.")
    print("Please install gfortran:")
    print("  Linux: sudo apt-get install gfortran")
    print("  macOS: brew install gcc")
    print("  Windows: conda install m2w64-toolchain")
    return False


# Check gfortran availability at setup time (skip in docs mode)
if not DOCS_BUILD_MODE:
    check_gfortran()
else:
    print("📚 Skipping gfortran check in documentation build mode")


# gridfill extensions now handled by meson.build in src/skyborn/gridfill/


class MesonBuildExt(build_ext):
    """Custom build extension to handle meson builds for Fortran modules"""

    def run(self):
        """Run the build process"""
        print("DEBUG: MesonBuildExt.run() called")
        # Build meson modules first
        self.build_meson_modules()
        # Then run the standard build_ext
        super().run()

    def build_meson_modules(self):
        """Build modules that use meson (like spharm)"""
        print("DEBUG: build_meson_modules() called")

        # Skip meson builds in documentation mode
        if DOCS_BUILD_MODE:
            print("📚 Documentation build mode - skipping meson module builds")
            return

        # Determine target directory based on --inplace flag
        if self.inplace:
            print("DEBUG: --inplace detected, building to source directory")
            # spharm_target = Path("src") / "skyborn" / "spharm"
        else:
            print("DEBUG: Building to build directory")
            # spharm_target = Path(self.build_lib) / "skyborn" / "spharm"

        # Auto-discover meson modules based on directory structure
        # Each module should have a meson.build file
        meson_modules = self._discover_meson_modules()

        for module in meson_modules:
            print(f"DEBUG: Processing module {module['name']}")
            if self.should_build_meson_module(module):
                print(f"DEBUG: Building module {module['name']} with meson")
                self.build_meson_module(module)
            else:
                print(f"DEBUG: Skipping module {module['name']} - no meson.build found")

    def should_build_meson_module(self, module):
        """Check if we should build this meson module"""
        meson_build_file = module["path"] / "meson.build"
        return meson_build_file.exists()

    def check_meson_available(self):
        """Check if meson and ninja are available"""
        try:
            # Check meson
            result = subprocess.run(
                ["meson", "--version"], capture_output=True, text=True, timeout=10
            )
            if result.returncode != 0:
                return False, "meson not found"

            meson_version = result.stdout.strip()
            print(f"Found meson version: {meson_version}")

            # Check ninja
            result = subprocess.run(
                ["ninja", "--version"], capture_output=True, text=True, timeout=10
            )
            if result.returncode != 0:
                return False, "ninja not found"

            ninja_version = result.stdout.strip()
            print(f"Found ninja version: {ninja_version}")

            return True, None

        except (
            subprocess.TimeoutExpired,
            subprocess.SubprocessError,
            FileNotFoundError,
        ) as e:
            return False, str(e)

    def build_meson_module(self, module):
        """
        Build a meson module using the meson build system.
        """
        print(f"Building {module['name']} with meson build system...")

        # Check if meson and ninja are available
        meson_available, error_msg = self.check_meson_available()
        if not meson_available:
            print(f"ERROR: Meson build tools not available: {error_msg}")
            print("Please install meson and ninja:")
            print("  pip install meson ninja")
            print("  or: conda install meson ninja")
            raise RuntimeError(
                f"Meson build tools required but not available: {error_msg}"
            )

        module_path = module["path"]
        # Use build subdirectory as specified in requirements
        build_dir = module_path / "build"

        try:
            # Clean build directory
            if build_dir.exists():
                print(f"Cleaning existing build directory: {build_dir}")
                shutil.rmtree(build_dir)

            # Setup build directory
            build_dir.mkdir(parents=True, exist_ok=True)

            # Configure meson build with custom install directory for wheel builds
            print(f"Configuring meson build in {build_dir} (cwd={module_path})")

            setup_cmd = [
                "meson",
                "setup",
                "build",  # build directory inside module_path
                ".",  # source is current directory (module_path)
                "--buildtype=release",
                "-Db_lto=true",
            ]

            # For wheel builds, configure custom install directory
            if not self.inplace and hasattr(self, "build_lib") and self.build_lib:
                # Tell meson to install to our build directory instead of system
                build_lib_path = Path(self.build_lib).resolve()
                setup_cmd.extend(
                    [
                        f"--python.purelibdir={build_lib_path}",
                        f"--python.platlibdir={build_lib_path}",
                    ]
                )
                print(f"DEBUG: Configuring meson to install to: {build_lib_path}")

            print(f"Running: {' '.join(setup_cmd)} (cwd={module_path})")

            # Set up environment for conda gfortran across all platforms
            env = os.environ.copy()
            import platform

            conda_prefix = env.get("CONDA_PREFIX", "")
            if conda_prefix:
                system = platform.system()
                current_path = env.get("PATH", "")

                if system == "Windows":
                    # Windows conda environment setup
                    conda_bin = os.path.join(conda_prefix, "bin")
                    conda_library_bin = os.path.join(conda_prefix, "Library", "bin")
                    env["PATH"] = f"{conda_bin};{conda_library_bin};{current_path}"
                    print(
                        f"Enhanced PATH for Windows conda environment: {conda_prefix}"
                    )

                elif system in ["Linux", "Darwin"]:
                    # Linux and macOS conda environment setup
                    conda_bin = os.path.join(conda_prefix, "bin")
                    # On Unix-like systems, use colon separator and prepend to PATH
                    env["PATH"] = f"{conda_bin}:{current_path}"

                    # Add lib directory to LD_LIBRARY_PATH (Linux) or DYLD_LIBRARY_PATH (macOS)
                    conda_lib = os.path.join(conda_prefix, "lib")
                    if system == "Linux":
                        current_lib_path = env.get("LD_LIBRARY_PATH", "")
                        env["LD_LIBRARY_PATH"] = (
                            f"{conda_lib}:{current_lib_path}"
                            if current_lib_path
                            else conda_lib
                        )
                        print(
                            f"Enhanced PATH and LD_LIBRARY_PATH for Linux conda environment: {conda_prefix}"
                        )
                    else:  # macOS
                        current_lib_path = env.get("DYLD_LIBRARY_PATH", "")
                        env["DYLD_LIBRARY_PATH"] = (
                            f"{conda_lib}:{current_lib_path}"
                            if current_lib_path
                            else conda_lib
                        )
                        print(
                            f"Enhanced PATH and DYLD_LIBRARY_PATH for macOS conda environment: {conda_prefix}"
                        )

                else:
                    print(
                        f"Warning: Unknown platform {system}, using basic conda PATH setup"
                    )
                    conda_bin = os.path.join(conda_prefix, "bin")
                    env["PATH"] = f"{conda_bin}:{current_path}"

            subprocess.run(setup_cmd, cwd=str(module_path), check=True, env=env)

            # Build with ninja (run relative to module_path, target 'build')
            print(f"Building with ninja in {build_dir} (cwd={module_path})")
            build_cmd = ["ninja", "-C", "build"]

            print(f"Running: {' '.join(build_cmd)} (cwd={module_path})")
            result = subprocess.run(
                build_cmd,
                cwd=str(module_path),
                check=True,
                capture_output=True,
                text=True,
            )

            if result.stdout:
                print("Build output:", result.stdout)
            if result.stderr:
                print("Build warnings/errors:", result.stderr)

            # Install using meson (this will handle the path configuration we set up)
            if not self.inplace and hasattr(self, "build_lib") and self.build_lib:
                print(f"Installing meson build outputs to {self.build_lib}")
                install_cmd = ["meson", "install", "-C", "build", "--only-changed"]
                print(f"Running: {' '.join(install_cmd)} (cwd={module_path})")

                try:
                    install_result = subprocess.run(
                        install_cmd,
                        cwd=str(module_path),
                        check=True,
                        capture_output=True,
                        text=True,
                        env=env,
                    )
                    if install_result.stdout:
                        print("Install output:", install_result.stdout)
                except subprocess.CalledProcessError as e:
                    print(f"ERROR: Meson install failed: {e}")
                    print(
                        "This should not happen with the new setup. Please check meson configuration."
                    )
                    raise
            else:
                print("Inplace build - extensions handled by meson custom_target")

            print(f"Meson build for {module['name']} completed successfully!")

            self._built_modules = getattr(self, "_built_modules", set())
            self._built_modules.add(module["name"])

        except (subprocess.CalledProcessError, RuntimeError, FileNotFoundError) as e:
            print(f"ERROR: Meson build failed for {module['name']}: {e}")
            if isinstance(e, subprocess.CalledProcessError):
                print(f"Command failed with exit code: {e.returncode}")
                if hasattr(e, "stdout") and e.stdout:
                    print("Stdout:", e.stdout)
                if hasattr(e, "stderr") and e.stderr:
                    print("Stderr:", e.stderr)
            raise  # Re-raise the exception since we're not using f2py fallback

    def _discover_meson_modules(self):
        """
        Auto-discover meson modules by recursively looking for meson.build files
        in skyborn subpackages and their subdirectories.
        """
        modules = []
        skyborn_src = Path("src") / "skyborn"

        def _find_meson_builds(base_path, relative_path=""):
            """Recursively find meson.build files"""
            for subdir in base_path.iterdir():
                if subdir.is_dir():
                    meson_file = subdir / "meson.build"
                    if meson_file.exists():
                        if relative_path:
                            module_name = f"{relative_path}.{subdir.name}"
                        else:
                            module_name = subdir.name
                        print(f"DEBUG: Discovered meson module: {module_name}")
                        modules.append(
                            {
                                "name": module_name,
                                "path": subdir,
                            }
                        )
                    # Continue searching in subdirectories
                    new_relative_path = (
                        f"{relative_path}.{subdir.name}"
                        if relative_path
                        else subdir.name
                    )
                    _find_meson_builds(subdir, new_relative_path)

        # Start recursive search from skyborn root
        _find_meson_builds(skyborn_src)

        return modules


class CustomDevelop(develop):
    """Custom develop command that builds meson modules"""

    def run(self):
        # Build meson modules in develop mode
        self.run_command("build_ext")
        super().run()


class CustomInstall(install):
    """Custom install command that ensures meson modules are built"""

    def run(self):
        # Ensure meson modules are built before install
        self.run_command("build_ext")
        super().run()


# Configuration for mixed build
setup_config = {
    "cmdclass": {
        "build_ext": MesonBuildExt,
        "develop": CustomDevelop,
        "install": CustomInstall,
    },
    # Add extensions for dummy (Windows compatibility) only
    # gridfill extensions now handled by meson.build
    "ext_modules": [
        Extension("skyborn._dummy", sources=["src/skyborn/_dummy.c"], optional=True)
    ],
}

if __name__ == "__main__":
    setup(**setup_config)



"""Example data utilities."""

from __future__ import absolute_import

import os


def example_data_path(filename=None):
    """Return the full path to the example data directory."""
    if filename is None:
        filename = ""
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "example_data", filename
    )


__all__ = ["example_data_path"]
